\ Information
\
\ 3-point tracking calculator for LC tuned circuits
\
\ Based on equations published in 'Electronic Valves' vol.4
\ Philips 1950
\
\ Compile with DX-Forth:  FORTH-F - INCLUDE TRACK.F BYE
\

empty forth definitions decimal
application

: TITLE ." TRACK version 0.3  29-Jul-13" cr ;

cr .( Compiling: )  title

\ Dependencies
cr .( Requires F.R from FPOUT package and 79-NOT. If you don't )
cr .( have them, enable the equate in source file. ) cr cr

0 [if]

  s" MAX-FLOAT-DIGITS" environment? [if] ( enhanced REPRESENT)
    drop  s" FPOUT34.F" included
  [else]
    s" FPOUT22.F" included
  [then]

: NOT  0= ;

[then]

: HELP  ( -- )
  cr ." 3-Point Tracking for LC Tuned Circuits" cr
  cr ." Given 3 capacitor and frequency values, calculate"
  cr ." inductance and capacitance required for tracking"
  cr ." at those points."
  cr ;

\ Display 'r' with 'prec' significant digits max
\ right-justified in a field of 'width' characters
: FP.R ( r prec width -- )
  >r set-precision -1 r> f.r ;


\ Variables
variable MODE               \ 0=low L/C, 1=high L/C

fvariable (C1)              \ capacitor 1
fvariable (C2)              \ capacitor 2
fvariable (C3)              \ capacitor 3

fvariable (F1)              \ frequency 1
fvariable (F2)              \ frequency 2
fvariable (F3)              \ frequency 3

3.14159265358979324E        \ PI
fdup f+  fconstant 2PI      \ PI*2

: C1  ( f: -- r )  (c1) f@ ;
: C2  ( f: -- r )  (c2) f@ ;
: C3  ( f: -- r )  (c3) f@ ;

: F1  ( f: -- r )  (f1) f@ ;
: F2  ( f: -- r )  (f2) f@ ;
: F3  ( f: -- r )  (f3) f@ ;

\ Equation factors
: S  ( f: -- r )  1e  2pi f1 f*  fdup f*  f/ ;
: E  ( f: -- r )  1e  2pi f2 f*  fdup f*  f/ ;
: N  ( f: -- r )  1e  2pi f3 f*  fdup f*  f/ ;

: E-N  ( f: -- r )  e n f- ;
: N-S  ( f: -- r )  n s f- ;
: S-E  ( f: -- r )  s e f- ;

: C1-C2  ( f: -- r )  c1 c2 f- ;
: C2-C3  ( f: -- r )  c2 c3 f- ;
: C3-C1  ( f: -- r )  c3 c1 f- ;

: C2*C3  ( f: -- r )  c2 c3  f* ;
: C1*C3  ( f: -- r )  c1 c3  f* ;
: C1*C2  ( f: -- r )  c1 c2  f* ;

\ Padding capacitor
: CP1  ( f: -- r )
  e-n c2*c3  f*
  n-s c1*c3  f*  f+
  s-e c1*c2  f*  f+
  e-n c1  f*
  n-s c2  f*  f+
  s-e c3  f*  f+
  f/ ;

\ Parallel capacitor
: CO1  ( f: -- r )
  e-n c2*c3  f*
  n-s c1*c3  f*  f+
  s-e c1*c2  f*  f+
  s-e e-n n-s  f* f*
  f/
  s e-n c2*c3  f* f*
  e n-s c1*c3  f* f*  f+
  n s-e c1*c2  f* f*  f+
  c1-c2 c2-c3 c3-c1  f* f*
  f/
  f* ;

\ Inductor
: L1  ( f: -- r )
  s-e e-n n-s  f* f*
  c1-c2 c2-c3 c3-c1  f* f*
  f*
  e-n c2*c3  f*
  n-s c1*c3  f*  f+
  s-e c1*c2  f*  f+
  fdup f*
  f/ ;

\ Padding capacitor
: CP2  ( f: -- r )
  s-e e-n n-s  f* f*
  e-n c1  f*
  n-s c2  f*  f+
  s-e c3  f*  f+
  f/
  c1-c2 c2-c3 c3-c1  f* f*
  s e-n c1  f* f*
  e n-s c2  f* f*  f+
  n s-e c3  f* f*  f+
  f/
  f* ;

\ Parallel capacitor
: CO2  ( f: -- r )
  s e-n c2*c3  f* f*
  e n-s c1*c3  f* f*  f+
  n s-e c1*c2  f* f*  f+
  s e-n c1  f* f*
  e n-s c2  f* f*  f+
  n s-e c3  f* f*  f+
  f/ ;

\ Inductor
: L2  ( f: -- r )
  s e-n c1  f* f*
  e n-s c2  f* f*  f+
  n s-e c3  f* f*  f+
  fdup f*
  s-e e-n n-s  f* f*
  c1-c2 c2-c3 c3-c1  f* f*
  f*
  f/ ;

\ padding capacitor
: CP  ( f: -- r )
  mode @ if  cp2  else  cp1  then
  1e-12 f* ;  \ re-scale

\ parallel capacitor
: CO  ( f: -- r )
  mode @ if  co2  else  co1  then
  1e-12 f* ;  \ re-scale

\ inductor
: LP  ( f: -- r )
  mode @ if  l2  else  l1  then ;

\ Scale frequency and display right-justified
: FR.R  ( n1 n2 -- ) ( f: r -- )
  2>r
  fdup fabs 1e3 f< if
    s" Hz"
  else
    fdup fabs 1e6 f< if
      1e3 f/  s" KHz"
    else
      1e6 f/  s" MHz"
    then
  then
  2r> 2swap 2>r
  fp.r  space  2r> type ;

\ Scale inductor and display right-justified
: L.R  ( n1 n2 -- ) ( f: r -- )
  2>r
  fdup fabs 1e f< not if
    s" H"
  else
    fdup fabs 1e-3 f< not if
      1e3 f*  s" mH"
    else
      1e6 f*  s" uH"
    then
  then
  2r> 2swap 2>r
  fp.r  space  2r> type ;

\ Scale capacitor and display right-justified
: C.R  ( n1 n2 -- ) ( f: r -- )
  2>r
  fdup fabs 1e-6 f< not if
    1e6 f*  s" uF"
  else
    fdup fabs 1e-9 f< not if
      1e9 f*  s" nF"
    else
      1e12 f*  s" pF"
    then
  then
  2r> 2swap 2>r
  fp.r  space  2r> type ;

\ Display results for current mode
: .RESULT  ( -- )
  cr ." Cp = "  cp  3 6 c.r
  cr ." Co = "  co  3 6 c.r
  cr ." L  = "  lp  3 6 l.r
  cr ;

\ Display results
: SHOWRESULTS  ( -- )
  cr ." Low L/C circuit:"   0 mode !  .result
  cr ." High L/C circuit:"  1 mode !  .result ;

\ Input floating-point number with prompt string
: ENTER  ( addr u -- ) ( f: -- r )
  cr type  ."  ? "  pad  dup 80 accept
  >float  0= if  r> drop  0 exit  then
  fdup  f0= if  fdrop  r> drop  0 exit  then ;

\ Enter input values
: INPUTVALUES  ( -- flag )
  s" C1 pF " enter  (c1) f!
  s" C2 pF " enter  (c2) f!
  s" C3 pF " enter  (c3) f!
  s" F1 MHz" enter  (f1) f!
  s" F2 MHz" enter  (f2) f!
  s" F3 MHz" enter  (f3) f!  cr
  -1 ;

\ Run application
: RUN  ( -- )
  help
  precision fdp @ 2>r  fdp off
  begin
    inputvalues
  while
    showresults
  repeat  cr
  2r> fdp ! set-precision ;

\ PROGRAM
: PROGRAM  ( -- )
  cr title              \ show program name
  run ;                 \ run application


cr .( Save to disk? )  y/n
[if]  turnkey program  TRACK  [then]

